#!/usr/bin/perl -w

# This script creates an animation from a Pyxida log server log.
#
# Todos:
#
#   * add proper conversion routing from 4D -> 3D
#   * add a more intelligent range calculation when choosing what to display
#

use strict;

use Getopt::Std;
use IPC::Open2;
use Statistics::Descriptive;
use Aux;

my $framePeriod = 120; 									# frame period in sec
my $debug = 0;
my $frameFormat = "png";
my $codec = "ffv1";
my $fps = 25;

my $rangePercentile = 99;
my $alpha = 0.01;

my $showShadows = 0;
my $zoffset = 100;

######################################################################

my %params = ();
my %pos = ();
getopts("f:dSAc:a:", \%params);

my $coordType;
$coordType = "S" if defined $params{'S'};
$coordType = "A" if defined $params{'A'};

&usage unless $coordType;

sub usage() {
	print << "EOF"; 
	usage: cat <pyxida-serverlog.log> | animateNC.pl [-f sec] [-d] <-S|-A>
	
	       -S          :              Use system-level coordinates
	       -A          :              Use application-level coordinates
	       -f          :              Frame perdiod in seconds
	       -d          :              Turn on debug mode
               -c          :              cn2rgb file
               -a          :              id2asn file

EOF
	die;
}

$framePeriod = $params{'f'} if defined $params{'f'};
$debug = 1 if defined $params{'d'};

######################################################################
# slurp color files

my %cn2color = ();
my %id2cn = ();
my $cn2color_file = 'data/animate/cn2rgb';
if (defined($params{'c'})) {
    $cn2color_file = $params{'c'};
}
my $id2cn_file = 'data/animate/id2asn';
if (defined($params{'a'})) {
    $id2cn_file = $params{'a'};
}

my %noColorFound = ();

&Aux::slurp_columns_to_hash_pair ($cn2color_file,\%cn2color, 0, 1, 3);
# hostname -> country
&Aux::slurp_column_to_hash_pair ($id2cn_file,\%id2cn, 3, 2);
# ip -> country
&Aux::slurp_column_to_hash_pair ($id2cn_file,\%id2cn, 4, 2);

######################################################################

my $startFrameTS;
my $frameCounter = 0;

my $xrange, my $yrange, my $zrange;
&clearRange();

my $s_xmin = 0;
my $s_xmax = 0;
my $s_ymin = 0;
my $s_ymax = 0;
my $s_zmin = 0;
my $s_zmax = 0;

local (*GPIN, *GPOUT);
open2(\*GPIN, \*GPOUT, "gnuplot"); 
#open GPOUT, ">/tmp/gp.test" or die ("Cannot open gp.test");
print STDERR "Launching gnuplot...\n" if ($debug);

while (my $line = <>) {
	if ($line =~ m/\w\w (\d+) \S+ a=(\S+) .+ sc2=\[(.+),(.+),(.+),h(.+)\] ac2=\[(.+),(.+),(.+),h(.+)\]/) {
		my ($ts, $addr) = ($1 / 1000, $2);
		my @sc = ($3, $4, $5, $6);
		my @ac = ($7, $8, $9, $10);
		my $dateTS = localtime($ts);
		print STDERR $line if ($debug);
		print STDERR "$dateTS $ts $addr " . join(",", @sc) . " " . join (",", @ac) . "\n" if ($debug);

		print STDERR "\nsc=" . join(" ", @sc) . "\n" if ($debug);
		print STDERR "ac=" . join(" ", @ac) . "\n\n" if ($debug);
				        			
		my @newCoord = ();				
		if ($coordType eq 'S') {
			@newCoord = @sc;
		} else {
			@newCoord = @ac;
		}		
		
		my $coordStr = join("", @newCoord);
		#print STDERR "$coordStr\n";
		if ($coordStr =~ m/,/) {
			print STDERR "Found bad coordinate. Skiping.\n" if ($debug);
			next;
		}

		@newCoord = &projectSimple(@newCoord);
		
		$pos{$addr} = [ @newCoord ];
		
		print STDERR "3D: " . join(" ", @{$pos{$addr}}) . "\n" if ($debug);
				
		&updateRange($pos{$addr});							
		
		$startFrameTS = $ts unless defined $startFrameTS;
		
		if ($ts - $startFrameTS > $framePeriod) {
			print STDERR "Outputting frame\n" if ($debug);
			$frameCounter++;
			&outputFrame($dateTS);
			%pos = ();
			&clearRange();
			$startFrameTS = $ts;
			print STDERR "Starting a new frame at " . $dateTS . "\n";			
		}		
	}			
}

close GPIN;
close GPOUT;

#warn ("Not generating animation");
#exit (0);

print STDERR "Generating animation...\n";
system "mencoder \"mf:///tmp/*.png\" -mf fps=$fps -o pyxida-anim-$coordType.avi -ovc lavc -lavcopts vcodec=$codec";
print STDERR "Deleting frames...\n";
system "rm /tmp/pyxida-anim-*.$frameFormat\n";
print STDERR "Done.\n";

sub outputFrame() {
	my $title = shift;

	my $filenameCount = &fixedNum($frameCounter);
	print GPOUT "set terminal $frameFormat size 1024,768\n";
	print GPOUT "set output \'/tmp/pyxida-anim-$filenameCount.$frameFormat\'\n";
	print GPOUT "set title \'$title\'\n";
	
	print GPOUT << "EOF";
set grid nopolar
set grid xtics nomxtics ytics nomytics noztics nomztics nox2tics nomx2tics noy2tics nomy2tics nocbtics nomcbtics
set grid layerdefault

EOF

	my $xmin = $xrange->percentile(100 - $rangePercentile);
	$xmin = 0 unless defined $xmin;	
	my $xmax = $xrange->percentile($rangePercentile);
	$xmax = 0 unless defined $xmax;
	my $ymin = $yrange->percentile(100 - $rangePercentile);
	$ymin = 0 unless defined $ymin;
	my $ymax = $yrange->percentile($rangePercentile);
	$ymax = 0 unless defined $ymax;
	my $zmin = $zrange->percentile(100 - $rangePercentile);
	$zmin = 0 unless defined $zmin;
	my $zmax = $zrange->percentile($rangePercentile);
	$zmax = 0 unless defined $zmax;
	
	$s_xmin = $xmin unless defined $s_xmin;
	$s_xmax = $xmax unless defined $s_xmax;
	$s_ymin = $ymin unless defined $s_ymin;
	$s_ymax = $ymax unless defined $s_ymax;
	$s_zmin = $zmin unless defined $s_zmin;
	$s_zmax = $zmax unless defined $s_zmax;	
	
	$s_xmin = (1 - $alpha) * $s_xmin + ($alpha * $xmin);
	$s_xmax = (1 - $alpha) * $s_xmax + ($alpha * $xmax); 
	$s_ymin = (1 - $alpha) * $s_ymin + ($alpha * $ymin); 
	$s_ymax = (1 - $alpha) * $s_ymax + ($alpha * $ymax); 
	$s_zmin = (1 - $alpha) * $s_zmin + ($alpha * $zmin); 
	$s_zmax = (1 - $alpha) * $s_zmax + ($alpha * $zmax); 	 
				
	#for (my $i = 0; $i < 100; $i++) {
	#	print "$i " . $xrange->percentile($i) . "\n";
	#}

	print STDERR "[$s_xmin:$s_xmax] [$s_ymin:$s_ymax] [$s_zmin:$s_zmax]\n" if ($debug); 	
		
	print GPOUT "set xrange [$s_xmin:$s_xmax]\n";
	print GPOUT "set yrange [$s_ymin:$s_ymax]\n"; 	
	print GPOUT "set zrange [" . ($s_zmin - $zoffset) . ":$s_zmax]\n"; 	

	print GPOUT 'rgb(r,g,b) = 65536 * int(r) + 256 * int(g) + int(b)'."\n";
	# XXX where should shadow be
	print GPOUT "flat(p) = ".($s_zmin - $zoffset + 10)."\n";
	
	#print GPOUT "set pointsize 1.5\n";
	
	#print GPOUT "splot '-' notitle\n";
	# notitle?
	print GPOUT "splot '-' " . 'using 1:2:3:(rgb($4,$5,$6)) notitle with points pt 7 lc rgb variable';
	if ($showShadows) {
		print GPOUT ", '-' ".'using 1:2:(flat($3)) '."notitle with points lc \"#CCCCCC\"\n";
	} else {
		print GPOUT "\n";
	}

	my $shadow = '';
	foreach my $addr (keys %pos) {		
	    print STDERR "$addr\n" if ($debug);
	    my $points = join(" ", @{$pos{$addr}});
	    $shadow .= "$points\n";
	    $addr =~ s/\///g;
	    $addr =~ s/:\d*//g;
	    $addr = lc($addr);
	    my $color = &getColor($addr);
	    print GPOUT "$points $color\n";
	    #print "$points $color\n";
	}
	print GPOUT "e\n";
	print GPOUT "$shadow"."e\n" if ($showShadows);
}

sub updateRange() {
	my $coords = shift;
	print STDERR "Updating range $$coords[0]\n" if ($debug);
			
	$xrange->add_data($$coords[0]);	
	$yrange->add_data($$coords[1]);	
	$zrange->add_data($$coords[2]);	
}

sub clearRange() {
	$xrange = new Statistics::Descriptive::Full->new();
	$yrange = new Statistics::Descriptive::Full->new();
	$zrange = new Statistics::Descriptive::Full->new();
}

sub fixedNum() {
	my $num = shift;
	my $result = sprintf("%05d", $num);
	return $result;
}

#sub projectClever4D() {
#	my @coord4D = @_;
#	print STDERR "4D: $coord4D[0] $coord4D[1] $coord4D[2] $coord4D[3]\n" if ($debug);
#	
#	# x' = k * (x / w)
#   # y' = k * (y / w)
#   # z' = k * (z / w)
#    
#    my @coord3D = ();
#    if ($coord4D[3] == 0) {
#    	@coord3D = (0, 0, 0);	
#    } else {
#		@coord3D = ( $k * $coord4D[0] / $coord4D[3], $k * $coord4D[1] / $coord4D[3], $k * $coord4D[2] / $coord4D[3] );
#    }	
#	return @coord3D;
#}

sub projectSimple() {
	my @coord3Dh = @_;
	print STDERR "3D+h: $coord3Dh[0] $coord3Dh[1] $coord3Dh[2] $coord3Dh[3]\n" if ($debug);
	my @coord3D = ($coord3Dh[0], $coord3Dh[1], $coord3Dh[2]);
	return @coord3D;
}


######################################################################

sub getColor {
    my ($host) = @_;

    #print "h $host\n";
    #print "id $id2cn{$host}\n";
    #print "cn $cn2color{$id2cn{$host}}\n";

    #exit (0);

    my $color = '';

    # hack around unknown hosts
    my $cn = '';
    if (defined($id2cn{$host})) {
		$cn = $id2cn{$host};
    } else {

	# try to match the country 
	if ($host =~ /\.(\w*)$/) {
	    $cn = $1;
	    if ($cn =~ /edu/i) { $cn = 'US'; }
	    $cn = uc($cn);
	}
    }

    if (defined($cn2color{$cn})) {
		$color = $cn2color{$cn};
    } else {

	if (!defined($noColorFound{$host})) {
	    print "no color for $host\n";
	    $noColorFound{$host} = 1;
	}
	$color = '00 00 00';
    }
    return $color;
}
