#!/usr/bin/perl

use strict;
use Getopt::Std;
use Frontier::Client;
use Data::Dumper;

my $PORT=55504;
my $API_PORT=55501;
my $USE_HEIGHT = 1;
my $DEBUG = 0;

my $usage = "query.pl [options] url\n";
$usage .=   " -l: get local coordinate of node running at url\n";
$usage .=   " -s subset.dat: file containing list of remote nodes, one per line\n";
$usage .=   " -r remote-node: get remote overlay node coordinate\n";
$usage .=   " -p proxy-node: get proxy coord that is stored on contacted node\n";
$usage .=   " -a nodeA -b nodeB: estimate latency btw overlay nodes A and B\n";
$usage .=   " -d debug\n";

$usage .=   "\n";
$usage .=   "query.pl -l http://righthand.eecs.harvard.edu\n";
$usage .=   "query.pl -a sb02.eecs.harvard.edu -b sb03.eecs.harvard.edu http://sb01.eecs.harvard.edu\n";
$usage .=   "Default port numbers can be overridden with \":port\" syntax\n";
my %para = ();
getopts('lr:a:b:s:d', \%para);

############################################################

if (defined($para{'d'})) {
    $DEBUG = 1;
}

if ($#ARGV < 0) {
    die ("Missing arguments\n$usage");
}
my $url = $ARGV[0];

if ($url !~ /^http:\/\//) {
    $url = 'http://'.$url;
}

if ($url =~ /^http:\/\/[\w+\.-]+$/) {
    $url .= ":$API_PORT/xmlrpc"
}

if ($url =~ /^http:\/\/[\w+\.-]+:\d+$/) {
		$url .= "/xmlrpc";
}

if ($url !~ /^http:\/\/[\w+\.-]+:\d+\/xmlrpc$/) {
    die ("word: $url\n$usage\n");
}

############################################################
#warn "url $url\n";
my $server = Frontier::Client->new(url => "$url", debug => $DEBUG);

if (defined($para{'l'})) {
    my $coordArray = $server->call("pyxida.getLocalCoord");
    if ($DEBUG) {
	print Dumper($coordArray);
    }
    my $coord = &parse_coord ($coordArray);
    print "$coord\n";

} elsif (defined($para{'s'})) {
    my %remoteNodes = ();
    open IN, "$para{'s'}" or die ("Cannot open file $para{'s'}\n");
    while (my $line = <IN>) {
	$line =~ s/\s//g;
	next if ($line =~ /\#/);
	if ($line =~ /\w+/) {
	    my $node = &addPort ($line,$PORT);
	    $remoteNodes{$node} = 1;
	}
    }
    close IN;    
    my $node2coord = $server->call("pyxida.getRemoteCoords", \%remoteNodes);
    if ($DEBUG) {
	print Dumper($node2coord);
    }
    foreach my $node (keys %$node2coord) {
	my $coordArray = $node2coord->{$node};
	if ($node ne 'result') {
	    my $coord = &parse_coord ($coordArray);
	    print "$node $coord\n";
	}
    }
    if ((defined($node2coord->{'result'})) &&
	($node2coord->{'result'} ne 'Success')) {
	warn "$node2coord->{'result'}\n";
    }

} elsif (defined($para{'r'})) {
    my $remoteNode = &addPort ($para{'r'},$PORT);
    my $resHash = $server->call("pyxida.getRemoteCoord", $remoteNode);
    if ($DEBUG) {
	print Dumper($resHash);
    }
    if (defined($resHash->{'result'})) {
	if (($resHash->{'result'} eq 'Success') &&
	    (defined($resHash->{'coord'}))) {
	    my $coord = &parse_coord ($resHash->{'coord'});
	    print "$coord\n";
	} else {
	    warn "Failure: $resHash->{'result'}\n";
	}
    } else {
	warn "Failure: cause unknown\n";
    }
} elsif (defined($para{'a'}) && defined($para{'b'})) {
    my $nodeA = &addPort($para{'a'},$PORT);
    my $nodeB = &addPort($para{'b'},$PORT);

    my $distance = $server->call("pyxida.estimateRTT", $nodeA, $nodeB);
    if ($DEBUG) {
	print Dumper($distance);
    }
    if (defined($distance)) {
	$distance = sprintf ("%.3f", $distance);
	print "$distance\n";
    }
} else {
    die ("Missing arguments\n$usage");
}

sub parse_coord {
    my ($coord) = @_;
    my $out = '[';
    for (my $i = 0; $i <= $#$coord; $i++) {
        my $element = sprintf ("%.3f", $coord->[$i]);
	if ($i == $#$coord && $USE_HEIGHT) {
	    $out .= 'h';
	}
        $out .= $element;
        if ($i != $#$coord) {
            $out .= ' ';
        }
    }
    $out .= ']';
    return $out;
}

sub addPort {
    my ($host,$port) = @_;
    if ($host !~ /[\w]+:\d+/) {
	$host .= ":$port";
    }
    return $host;
}
